/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.afm;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataUtils;
import de.ozgcloud.eingang.common.formdata.ServiceKonto;
import de.ozgcloud.eingang.semantik.common.ServiceKontoFactory;

class AfmHeaderMapperTest {

	@Spy
	@InjectMocks
	private AfmHeaderMapper mapper;
	@Mock
	private ServiceKontoFactory serviceKontoFactory;

	@DisplayName("Parse form data")
	@Nested
	class TestParseFormData {

		@DisplayName("map form header")
		@Nested
		class TestMapFormHeader {

			private static final FormData FORM_DATA = FormData.builder().formData(AfmHeaderTestFactory.createFormDataMap()).build();

			@Test
			void shouldKeepHeader() {
				var parsedFormData = parseFormData();

				assertThat(parsedFormData.getFormData().get(AfmHeaderMapper.HEADER_FIELD)).isNotNull();
			}

			@DisplayName("fields")
			@Nested
			class TestFields {

				@Test
				void shouldMapRequestId() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getRequestId()).isEqualTo(AfmHeaderTestFactory.ID);
				}

				@Test
				void shouldMapVorgangNummer() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getVorgangNummer()).isEqualTo(AfmHeaderTestFactory.ID);
				}

				@Test
				void shouldMapCreatedAt() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getCreatedAt()).isEqualTo(AfmHeaderTestFactory.TIMESTAMP);
				}

				@Test
				void shouldMapFormId() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getFormId()).isEqualTo(AfmHeaderTestFactory.FORM_ID);
				}

				@Test
				void shouldMapFormName() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getFormName()).isEqualTo(AfmHeaderTestFactory.FORM);
				}

				@Test
				void shouldMapSender() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getSender()).isEqualTo(AfmHeaderTestFactory.SENDER);
				}

				@Test
				void shouldSetFormEngineName() {
					var parsedFormData = parseFormData();

					assertThat(parsedFormData.getHeader().getFormEngineName()).isEqualTo(AfmHeaderMapper.AFM_FORMENGINE_NAME);
				}

				@DisplayName("service konto")
				@Nested
				class TestServiceKonto {

					@DisplayName("OSI")
					@Nested
					class TestOsiServiceKonto {
						@Test
						void shouldCallBuildServiceKontoIfPresent() {
							parseFormData();

							verify(serviceKontoFactory).buildOsiServiceKonto(any(), eq(FORM_DATA));
						}

						@Test
						void shouldNotCallBuildServiceKontoIfNotExists() {
							mapper.parseFormData(FormDataUtils.from(FORM_DATA).remove(AfmHeaderMapper.POSTFACH_NAME_ID).build());

							verify(serviceKontoFactory, never()).buildOsiServiceKonto(any(), any());
						}
					}

					@DisplayName("BayernID")
					@Nested
					class TestBayernId {

						@Mock
						private FormData formData;
						@Mock
						private ServiceKonto serviceKonto;

						@Test
						void shouldCallCreateBayernIdServiceKonto() {
							var formData = FormData.builder().formData(AfmHeaderTestFactory.createFormDataMapWithExtendedHeaders()).build();

							mapper.parseFormData(formData);

							verify(mapper).createBayernIdServiceKonto(formData);
						}

						@Test
						void shouldReturnServiceKonto() {
							when(serviceKontoFactory.createBayernIdServiceKonto(any())).thenReturn(Optional.of(serviceKonto));

							var parsedFormData = mapper.createBayernIdServiceKonto(formData);

							assertThat(parsedFormData).contains(serviceKonto);
						}

					}
				}
			}

			private FormData parseFormData() {
				return mapper.parseFormData(FORM_DATA);
			}
		}

		@DisplayName("remove mapped data")
		@Nested
		class TestRemoveMappedData {

			private final FormData formData = FormData.builder().formData(AfmHeaderTestFactory.createFormDataMap()).build();

			@Test
			void shouldRemoveRestResponseName() {
				var parsedFormData = parseFormData();

				assertThat(parsedFormData.getFormData().get(ServiceKontoFactory.REST_RESPONSE_NAME)).isNull();
			}

			private FormData parseFormData() {
				return mapper.parseFormData(formData);
			}
		}
	}

}
