package de.ozgcloud.eingang.semantik.enginebased.afm.intelliform;

import static de.ozgcloud.eingang.semantik.enginebased.afm.intelliform.IntelliFormRepresentationAdapter.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;

class IntelliFormRepresentationAdapterTest {

	private static final String DOCUMENT_TEST_KEY = "doc-test-key";
	private static final String DOCUMENT_TEST_VALUE = "doc-test-value";

	@Spy
	@InjectMocks
	IntelliFormRepresentationAdapter adapter;

	@Mock
	private XmlToJavaMapsMapper xmlToJavaMapsMapper;

	@DisplayName("find IntelliForm XML representation")
	@Nested
	class TestFindIntelliFormXmlRepresentation {

		@Mock
		private FormData formData;

		@DisplayName("should find first with IntelliForm schema")
		@Test
		void shouldFindFirstWithIntelliFormSchema() {
			var schemas = Stream.of(
					"<root xmlns=\"%s\" />".formatted(INTELLIFORM_TYPENAME),
					"<root xmlns:t=\"unknown\">first</root>",
					"<root>second</root>",
					"<root xmlns:t=\"%s\">third</root>".formatted(INTELLIFORM_TYPENAME),
					"<root xmlns:t=\"%s\">fourth</root>".formatted(INTELLIFORM_TYPENAME)
			).map(IntelliFormRepresentationAdapterTest.this::buildXMLDocument);
			doReturn(schemas).when(adapter).streamXMLDocumentsFromRepresentations(formData);

			var firstWithSchema = adapter.findIntelliFormXMLRepresentation(formData);

			assertThat(firstWithSchema).isPresent();
			var rootElement = firstWithSchema.get().getDocumentElement();
			assertThat(rootElement.getTextContent()).isEqualTo("third");
		}
	}

	@DisplayName("stream XML documents from representations")
	@Nested
	class TestStreamXmlDocumentsFromRepresentations {

		@Mock
		private Document document;

		@Mock
		private FormData formData;

		@Mock
		private InputStream inputStream;

		@DisplayName("with no representations")
		@Nested
		class TestWithNoRepresentations {
			@BeforeEach
			void mock() {
				when(formData.getRepresentations()).thenReturn(Collections.emptyList());
			}

			@DisplayName("should return empty")
			@Test
			void shouldReturnEmpty() {
				var documents = doExtract();

				assertThat(documents).isEmpty();
			}
		}

		@DisplayName("with a xml representation")
		@Nested
		class TestWithAXMLRepresentations {
			@Mock
			private IncomingFile otherRepresentation;

			@Mock
			private IncomingFile xmlRepresentation;

			@Mock
			private IncomingFile otherXMLRepresentation;

			@Mock
			private Document otherDocument;

			@Mock
			private InputStream otherInputStream;

			@BeforeEach
			void mock() {
				when(otherRepresentation.getContentType()).thenReturn("pdf");
				when(xmlRepresentation.getContentType()).thenReturn("xml");
				when(otherXMLRepresentation.getContentType()).thenReturn("xml");
				when(formData.getRepresentations()).thenReturn(List.of(otherRepresentation, xmlRepresentation, otherXMLRepresentation));

				when(xmlRepresentation.getContentStream()).thenReturn(inputStream);
				when(xmlToJavaMapsMapper.parseAsW3cDocument(inputStream)).thenReturn(document);

				when(otherXMLRepresentation.getContentStream()).thenReturn(otherInputStream);

			}

			@DisplayName("should include all successfully parsed documents")
			@Test
			void shouldIncludeAllSuccessfullyParsedDocuments() {
				when(xmlToJavaMapsMapper.parseAsW3cDocument(otherInputStream)).thenReturn(otherDocument);

				var documents = doExtract();

				assertThat(documents).isEqualTo(List.of(document, otherDocument));
			}

			@DisplayName("should skip documents with parse error")
			@Test
			void shouldSkipDocumentsWithParseError() {
				when(xmlToJavaMapsMapper.parseAsW3cDocument(otherInputStream)).thenThrow(new TechnicalException("some error"));

				var documents = doExtract();

				assertThat(documents).isEqualTo(List.of(document));
			}
		}

		private List<Document> doExtract() {
			return adapter.streamXMLDocumentsFromRepresentations(formData).toList();
		}
	}

	@DisplayName("is responsible")
	@Nested
	class TestIsResponsible {

		@Mock
		private FormData formData;

		@DisplayName("should be true with namespace attribute match")
		@Test
		void shouldBeTrueWithNamespaceAttributeMatch() {

			doReturn(Optional.of(buildXMLDocument("""
					<myForm xmlns:t="%s">
					</myForm>""".formatted(INTELLIFORM_TYPENAME))))
					.when(adapter).findIntelliFormXMLRepresentation(formData);

			var isResponsible = adapter.isResponsible(formData);

			assertThat(isResponsible).isTrue();
		}

		@DisplayName("should be false without xml document")
		@Test
		void shouldBeFalseWithoutXmlDocument() {
			doReturn(Optional.empty())
					.when(adapter).findIntelliFormXMLRepresentation(formData);

			var isResponsible = adapter.isResponsible(formData);

			assertThat(isResponsible).isFalse();
		}

	}

	@DisplayName("get form data entries from document")
	@Nested
	class TestGetFormDataEntriesFromDocument {

		private static final String OTHER_VENDOR_ID = "vendorId:other";
		private static final String MISSING_VENDOR_ID = "vendorId:missing";

		@Mock
		private Document document;

		private List<IncomingFileGroup> attachments;

		@BeforeEach
		void mock() {
			when(xmlToJavaMapsMapper.mapXmlToJavaMaps(document)).thenReturn(Map.of(
					DOCUMENT_TEST_KEY, DOCUMENT_TEST_VALUE,
					"Upload1", Map.of(
							"file", Map.of("id", IncomingFileGroupTestFactory.VENDOR_ID_XXX)
					),
					"Upload2", Map.of(
							"file", Map.of("id", OTHER_VENDOR_ID)
					),
					"Upload3", Map.of(
							"file", Map.of("id", "representationId")
					),
					"NoUpload1", Map.of(
							"file", ""
					),
					"NoUpload2", Map.of()
			));
			attachments = List.of(
					IncomingFileGroupTestFactory.create(),
					IncomingFileGroupTestFactory.createBuilder()
							.name("other")
							.files(List.of(
									IncomingFileTestFactory.createBuilder()
											.vendorId(MISSING_VENDOR_ID)
											.build(),
									IncomingFileTestFactory.createBuilder()
											.vendorId(OTHER_VENDOR_ID)
											.build()))
							.build());

		}

		@DisplayName("should prune upload element if is an attachment")
		@Test
		void shouldPruneUploadElementIfIsAnAttachment() {
			var entryKeys = obtainResidualEntryKeys();

			assertThat(entryKeys).containsExactlyInAnyOrder(DOCUMENT_TEST_KEY, "Upload3", "NoUpload1", "NoUpload2");
		}

		private List<String> obtainResidualEntryKeys() {
			return adapter.getFormDataEntriesFromDocument(document, attachments).map(Map.Entry::getKey).toList();
		}
	}

	@DisplayName("create header map")
	@Nested
	class TestCreateHeaderMap {

		private Document document;

		@DisplayName("with full fields")
		@Nested
		class TestWithFullFields {

			private Map<String, Object> headerMap;

			@BeforeEach
			void mock() {
				headerMap = Stream.concat(HEADER_ATTRIBUTE_NAMES.stream(), CUSTOM_HEADER_ATTRIBUTE_NAMES.stream())
						.collect(Collectors.toMap(name -> name, name -> LoremIpsum.getInstance().getName()));
				document = buildXMLDocument("<root %s/>".formatted(
								headerMap.entrySet().stream()
										.map(entry -> "%s=\"%s\"".formatted(entry.getKey(), entry.getValue()))
										.collect(Collectors.joining(" "))
						)
				);
			}

			@DisplayName("should include header field")
			@ParameterizedTest
			@MethodSource("headerNames")
			void shouldIncludeAllHeaderField(String headerName) {
				var resultHeaderMap = adapter.createHeaderMap(document);

				assertThat(resultHeaderMap).containsEntry(headerName, headerMap.get(headerName));
			}

			private static Stream<Arguments> headerNames() {
				return HEADER_ATTRIBUTE_NAMES.stream().map(Arguments::of);
			}

			@DisplayName("should include custom header field")
			@ParameterizedTest
			@MethodSource("customHeaderNames")
			void shouldIncludeAllCustomHeaderField(String headerName) {
				var resultHeaderMap = adapter.createHeaderMap(document);

				assertThat(resultHeaderMap).containsEntry(headerName, headerMap.get(headerName));
			}

			private static Stream<Arguments> customHeaderNames() {
				return CUSTOM_HEADER_ATTRIBUTE_NAMES.stream().map(Arguments::of);
			}
		}

		@DisplayName("with empty fields")
		@Nested
		class TestWithEmptyFields {
			@BeforeEach
			void mock() {
				document = buildXMLDocument("<root />");
			}

			@DisplayName("should include empty header field")
			@ParameterizedTest
			@MethodSource("headerNames")
			void shouldIncludeAllHeaderField(String headerName) {
				var resultHeaderMap = adapter.createHeaderMap(document);

				assertThat(resultHeaderMap).containsEntry(headerName, "");
			}

			private static Stream<Arguments> headerNames() {
				return HEADER_ATTRIBUTE_NAMES.stream().map(Arguments::of);
			}

			@DisplayName("should not include custom header field")
			@ParameterizedTest
			@MethodSource("customHeaderNames")
			void shouldNotIncludeAllCustomHeaderField(String headerName) {
				var resultHeaderMap = adapter.createHeaderMap(document);

				assertThat(resultHeaderMap).doesNotContainKey(headerName);
			}

			private static Stream<Arguments> customHeaderNames() {
				return CUSTOM_HEADER_ATTRIBUTE_NAMES.stream().map(Arguments::of);
			}
		}
	}

	@DisplayName("adapt by representation")
	@Nested
	class TestAdaptByRepresentation {

		@Mock
		private FormData formData;

		@DisplayName("should throw exception without xml document")
		@Test
		void shouldThrowExceptionWithoutXmlDocument() {
			doReturn(Optional.empty())
					.when(adapter).findIntelliFormXMLRepresentation(formData);

			assertThrows(TechnicalException.class, () ->
					adapter.adaptByRepresentations(formData)
			);
		}

		@DisplayName("with xml document")
		@Nested
		class TestWithXmlDocument {

			@Mock
			private IncomingFileGroup attachmentGroup;

			@Mock
			private IncomingFile xmlRepresentation;

			@Mock
			private IncomingFile otherRepresentation;

			@Mock
			private Document document;

			@Mock
			private Map<String, Object> headerFields;

			private List<IncomingFileGroup> attachments;
			private List<IncomingFile> representations;

			@BeforeEach
			void mock() {
				attachments = List.of(attachmentGroup);
				representations = List.of(xmlRepresentation, otherRepresentation);

				when(formData.getRepresentations()).thenReturn(representations);
				when(formData.getAttachments()).thenReturn(attachments);
				doReturn(Optional.of(document)).when(adapter).findIntelliFormXMLRepresentation(formData);

				doReturn(Map.of(DOCUMENT_TEST_KEY, DOCUMENT_TEST_VALUE).entrySet().stream())
						.when(adapter).getFormDataEntriesFromDocument(document, attachments);

				doReturn(headerFields)
						.when(adapter).createHeaderMap(document);
			}

			@DisplayName("should have document field")
			@Test
			void shouldHaveDocumentField() {
				var documentValue = obtainMappingByKey(DOCUMENT_TEST_KEY);

				assertThat(documentValue).isEqualTo(DOCUMENT_TEST_VALUE);
			}

			@DisplayName("should have header field")
			@Test
			void shouldHaveHeaderField() {
				var headerValue = obtainMappingByKey(HEADER_FIELD);

				assertThat(headerValue).isEqualTo(headerFields);
			}

			private Object obtainMappingByKey(String key) {
				var formDataResult = adapter.adaptByRepresentations(formData);

				return formDataResult.getFormData().get(key);
			}

			@DisplayName("should have attachments")
			@Test
			void shouldHaveAttachments() {
				var formDataResult = adapter.adaptByRepresentations(formData);

				assertThat(formDataResult.getAttachments()).isEqualTo(attachments);
			}

			@DisplayName("should have representations")
			@Test
			void shouldHaveRepresentations() {
				var formDataResult = adapter.adaptByRepresentations(formData);

				assertThat(formDataResult.getRepresentations()).isEqualTo(representations);
			}
		}
	}

	private Document buildXMLDocument(String xmlString) {
		try {
			return DocumentBuilderFactory.newInstance()
					.newDocumentBuilder()
					.parse(new InputSource(new StringReader(xmlString)));
		} catch (ParserConfigurationException | IOException | SAXException e) {
			throw new RuntimeException(e);
		}
	}

}
